import datetime
import gridfs
import json
import numpy as np
import os
import pandas as pd
from pymongo import MongoClient
import sys, traceback
from tqdm import tqdm
import time


class CreateDb():
    """
    Create database and populate with json data
    """
    def __init__(self, db_name="pecan_street"):

        def test_db_conn():
            '''
            Test mongodb server connection
            '''
            try:
                client = MongoClient('mongodb://localhost:27017/')
                client.server_info()
                print('now connected to mongodb at port 27017')
            except:
                print('Cannot connect to mongodb server')
                exit(-1)

        test_db_conn()
    
        def init_db_client(db_name):
            '''
            @param db_name: name of database, default [pecan_street]
            @return: database client
            '''
            try:
                client = MongoClient('mongodb://localhost:27017/')
                db = client[db_name]
                return db
            except:
                print('Cannot connect to mongodb server')
                exit(-1)

        self.db = init_db_client(db_name)
        print(self.db)
    
    @staticmethod
    def encode_charset(val):
        '''
        @param val: some instance of np.int64
        @return: int value if int64 else returns value back
        '''
        if isinstance(val, np.int64):
            val = int(val)
        return val

    @staticmethod
    def parse_date(datetime_str):
        '''
        removes unwanted syntax in time object
        @param datatime_str: time string
        @return string
        '''
        datetime_arr = datetime_str.split(' ')
        time_component = datetime_arr[1]
        time_component = time_component.split('-')
        time_component = time_component[0]
        return datetime_arr[0] + ' ' + time_component


    @staticmethod    
    def parse_meta_data(metadata, dataid_list, time_collection, file_path):
        '''
        Writes a json object represent the metadata for all the cities 'newyork', 'austin', 'california'
        for a specific time interval
        [metadata_field] -> [] || [metadata_field] -> ['dataid': {value of metadata_field}]
        @param metadata: pandas dataframe for metadata
        @param dataid_list: list of dataid's in dataframes for specific time collection
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @param file_path: root directory for metadata file
        @return: None
        '''
        if not os.path.isdir(file_path):
            print(file_path + ' is not a valid directory')
            return

        metadata_obj = {}

        # get column types for metadata
        metadata_type = metadata.dtypes
        
        # get building types and initialize multidict object
        building_types_list = list(set(metadata['building_type']))
        for building_type in building_types_list:
            metadata_obj[building_type] = []

        for col in metadata.columns.to_list():
            if col != 'dataid' and col != 'building_type':
                metadata_obj[col] = []
        
        # slice metadata
        metadata_slice = metadata.loc[dataid_list, :].reset_index()
        dataid_col = metadata_slice.columns.get_loc('dataid')

        for row in tqdm(range(metadata_slice.shape[0]), desc='Creating metafile', ncols=100):
            dataid = CreateDb.encode_charset(metadata_slice.iloc[row, dataid_col])
            for col in range(metadata_slice.shape[1]):
                value = metadata_slice.iloc[row, col]
                # set building type
                if metadata_slice.columns[col] == 'building_type':
                    metadata_obj[value].append(dataid)
                col_name = metadata_slice.columns[col]

                if col_name != 'dataid' and col_name != 'building_type':
                    if metadata_type[col_name] == np.float64 and not pd.isnull(value):
                        if value != 0.0:
                            if col_name == 'house_construction_year':
                                value = int(value)
                            metadata_obj[col_name].append({ 
                                str(dataid): CreateDb.encode_charset(value)
                                })
                    else:
                        if not pd.isnull(value):
                            metadata_obj[col_name].append(dataid)
                            
        with open(os.path.join(file_path, time_collection + '_meta.json'), 'w') as write_file:
            json.dump(metadata_obj, write_file)

    @staticmethod
    def create_multidict(df, city_collection):
        '''
        creates a multidictionary object that is the json schema of the file to be 
        inserted into tha database --> [city_collection][dataid]
        @param df: pandas dataframe for pecan street dataset
        @city_collection: name of city collection in database -- 'newyork', 'austin', 'california'
        @return: multidictionary object
        '''
        data_dict = {city_collection: {}}
        dataid_col = df.columns.get_loc('dataid')

        for row_i in tqdm(range(df.shape[0]), desc='Creating database schema', ncols=100):
            dataid = str(CreateDb.encode_charset(df.iloc[row_i, dataid_col]))

            if dataid not in data_dict[city_collection]:
                data_dict[city_collection][dataid] = []

        return data_dict

    @staticmethod    
    def serialize_data(df, time_col_name, time_collection, city_collection, file_path):
        '''
        Creates a serialized json file with the data set by the predefined schema in 
        create_multidict()
        @param df: pandas dataframe for pecan street dataset
        @param time_col_name: time column name in dataframe
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @param city_collection: name of city collection in database -- 'newyork', 'austin', 'california'
        @param file_path: root file
        '''

        if not os.path.isdir(file_path):
            print('no such directory ' + file_path)
            return

        dataid_col = df.columns.get_loc('dataid')
        time_col = df.columns.get_loc(time_col_name)

        # Create multidict obj
        data = CreateDb.create_multidict(df, city_collection)

        for row_i in tqdm(range(df.shape[0]), desc='Creating serialized json data', ncols=100):
            data_dict = {}

            time_str = CreateDb.parse_date(df.iloc[row_i, time_col])
            time_obj = datetime.datetime.strptime(time_str, '%Y-%m-%d %H:%M:%S')

            dataid = str(CreateDb.encode_charset(df.iloc[row_i, dataid_col]))

            for col_i in range(df.shape[1]):
                if col_i != dataid_col and col_i != time_col:
                    if not pd.isna(df.iloc[row_i, col_i]):
                        data_dict[df.columns[col_i]] = CreateDb.encode_charset(df.iloc[row_i, col_i])
                if col_i == time_col:
                    data_dict["time"] = str(time_obj.date()) + ' ' + str(time_obj.time())

            data[city_collection][dataid].append(data_dict)

        with open(os.path.join(file_path, time_collection + '_' + city_collection + '.json'), 'w') as write_file:
            json.dump(data, write_file)


    def single_insert(self, time_collection, **kwargs):
        '''
        Insert a single file into the database: metadata or file
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @param city_collection: name of city collection in database -- 'newyork', 'austin', 'california'
        @param Keyword Arguments: 
                                  data_file_dir: root file directories
                                  meta_file_dir: root directory for metadata file
        '''
        acceptable_keys = ['data_file_dir', 'meta_file_dir', 'city_collection']

        if len(kwargs.keys()) == 0:
            raise KeyError('expected either [data_file_dir] or [meta_file_dir] got None')

        if not all(key in acceptable_keys for key in kwargs.keys()):
            raise KeyError('argument keys must be either [data_file_dir] or [meta_file_dir]')

        if 'city_collection' in kwargs.keys() and 'data_file_dir' not in kwargs.keys():
            raise KeyError('[city_collection] provided, while [data_file_dir] is missing')

        if 'data_file_dir' in kwargs.keys() and 'city_collection' not in kwargs.keys():
            raise KeyError('[data_file_dir] provided, while [city_collection] is missing')


        try:
            fs = gridfs.GridFS(self.db, collection=time_collection)

            if 'data_file_dir' in kwargs.keys():
                city_collection = kwargs.get('city_collection')
                if os.path.isdir(kwargs['data_file_dir']):
                    with open(os.path.join(kwargs['data_file_dir'], time_collection + '_' + city_collection + '.json'), 'rb') as data:
                        fs.put(data, filename=city_collection)
                    print(city_collection + ' inserted into database')
            if 'meta_file_dir' in kwargs.keys():
                if os.path.isdir(kwargs['meta_file_dir']):
                    with open(os.path.join(kwargs['meta_file_dir'], time_collection + '_meta.json'), 'rb') as data:
                        fs.put(data, filename='metadata')
                    print('metadata inserted into database')
        except:
            traceback.print_exc(file=sys.stdout)
            print('Could not insert file into database')

    def batch_insert(self, file_dir):
        '''
        Insert all json files in the directory provided
        @param file_dir: root file directory
        '''

        if os.path.isdir(file_dir):
            files = [file for file in os.listdir(file_dir) if os.path.isfile(os.path.join(file_dir, file))]
            for file in tqdm(files, desc='Inserting all files', ncols=100):
                if file.split('.')[-1] == 'json':
                    time_collection = file.split('_', 3)[:3]
                    city_collection = file.split('_')[-1].split('.')[0]
                    time_collection = '_'.join(time_collection)
                    if city_collection == 'meta':
                        self.single_insert(time_collection=time_collection, meta_file_dir=file_dir)
                    else:
                        self.single_insert(
                            time_collection=time_collection, 
                            city_collection=city_collection, 
                            data_file_dir=file_dir)
        else:
            print('no such directory ' + file_dir ) 






import datetime
import gridfs
import json
import math
import os
import pandas as pd 
from pymongo import MongoClient
import sys, traceback
from tqdm import tqdm


class QueryDb():
    """
    Query local database with AND and OR db operations
    """
    def __init__(self, db_name="pecan_street"):

        def test_db_conn():
            '''
            Test mongodb server connection
            '''
            try:
                client = MongoClient('mongodb://localhost:27017/')
                client.server_info()
                print('now connected to mongodb at port 27017')
            except:
                print('Cannot connect to mongodb server')
                exit(-1)

        test_db_conn()
    
        def init_db_client(db_name):
            '''
            @param db_name: name of database, default [pecan_street]
            @return: database client
            '''
            try:
                client = MongoClient('mongodb://localhost:27017/')
                db = client[db_name]
                return db
            except:
                print('Cannot connect to mongodb server')
                exit(-1)

        self.db = init_db_client(db_name)
        self.meta_file='metadata'
        self.query_tolerance = 5  # query tolerance for 1 second time intervals

        self.building_types = ['Single-Family Home', 'Mobile Home', 'Town Home', 'Apartment', 'Sales']

        # Kewyord Arguments and their corresponding tol parameters
        self.tol_dict = {
            'total_amount_of_pv': 'total_pv_tol',
            'house_construction_year': 'year_tol',
            'total_square_footage' : 'total_sqft_tol',
            'first_floor_square_footage' : 'first_floor_sqft_tol',
            'second_floor_square_footage' : 'second_floor_sqft_tol',
            'third_floor_square_footage' : 'third_floor_sqft_tol',
            'half_floor_square_footage' : 'half_floor_sqft_tol',
            'lower_level_square_footage' : 'lower_level_sqft_tol'
            }
        
        self.file_params = ['air1', 'air2', 'air3', 'airwindowunit1',
                            'aquarium1', 'bathroom1', 'bathroom2', 'bedroom1', 'bedroom2',
                            'bedroom3', 'bedroom4', 'bedroom5', 'battery1', 'car1', 'car2',
                            'circpump1', 'clotheswasher1', 'clotheswasher_dryg1', 'diningroom1',
                            'diningroom2', 'dishwasher1', 'disposal1', 'drye1', 'dryg1', 'freezer1',
                            'furnace1', 'furnace2', 'garage1', 'garage2', 'grid', 'heater1',
                            'heater2', 'heater3', 'housefan1', 'icemaker1', 'jacuzzi1', 'kitchen1',
                            'kitchen2', 'kitchenapp1', 'kitchenapp2', 'lights_plugs1',
                            'lights_plugs2', 'lights_plugs3', 'lights_plugs4', 'lights_plugs5',
                            'lights_plugs6', 'livingroom1', 'livingroom2', 'microwave1', 'office1',
                            'outsidelights_plugs1', 'outsidelights_plugs2', 'oven1', 'oven2',
                            'pool1', 'pool2', 'poollight1', 'poolpump1', 'pump1', 'range1',
                            'refrigerator1', 'refrigerator2', 'security1', 'sewerpump1', 'shed1',
                            'solar', 'solar2', 'sprinkler1', 'sumppump1', 'utilityroom1',
                            'venthood1', 'waterheater1', 'waterheater2', 'wellpump1', 'winecooler1',
                            'leg1v', 'leg2v']
        print(self.db)

    def read_file(self, time_collection, city_collection):
        '''
        Read file from time collection in database
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @city_collection: name of city collection in database -- 'newyork', 'austin', 'california'
        @return: file from database
        '''
        try:
            fs = gridfs.GridFS(self.db, collection=time_collection)
            file = fs.get_last_version(filename=city_collection)
            return json.loads(file.read())
        except:
            print('file not found in database')

    def read_meta(self, time_collection):
        '''
        Read metadata file from time collection in database
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @return: metadata file from database
        '''
        try:
            fs = gridfs.GridFS(self.db, collection=time_collection)
            file = fs.get_last_version(filename=self.meta_file)
            return json.loads(file.read())
        except:
            print('metadata not in database')

    def check_tol_param(self, query_param, args):
        '''
        Check if dictionary query parameter has corresponding tol parameter
        @param query_param: list of query parameters
        @return: list item containing [True] || [False, string: missing tol parameter]
        '''
        for query in query_param:
            if isinstance(query, dict):
                query_key = [key for key in query.keys()][0]
                if self.tol_dict[query_key] not in args.keys():
                    return [False, self.tol_dict[query_key]]
        return [True]

    @staticmethod
    def stringify_list(multi_list):
        new_list = []
        for list_item in multi_list:
            temp_list = []
            for item in list_item:
                temp_list.append(str(item))
            new_list.append(temp_list)
        return new_list

    def get_dataids(self, query_param, metadata, args):
        '''
        Filters dataids based on the type of db_op and query parameters
        @param query_param: list of query parameters
        @param metadata: metadata for a specific time collection
        @param args: Keyword Arguments
        @return: list of dataids
        '''
        dataid_list = [metadata[query] for query in query_param if isinstance(query, str) and query != 'nest']

        if 'nest' in query_param:
            nests = []
            for val in metadata['number_of_nests']:
                for key in val.keys():
                    nests.append(key)
            dataid_list.append(nests)

        for query in query_param:
            if isinstance(query, dict):
                dataid = []
                query_key = [key for key in query.keys()][0]
                for values in metadata[query_key]:
                    for meta_key, meta_val in values.items():
                        if math.fabs(meta_val  - query[query_key]) <= args[self.tol_dict[query_key]]:
                            dataid.append(meta_key)
                dataid_list.append(dataid)

        if len(dataid_list) > 1:
            dataid_list = QueryDb.stringify_list(dataid_list)
            if args['db_op'] == 'AND':
                dataid = dataid_list[0]
                for i in range(1, len(dataid_list), 1):
                    dataid = set(dataid).intersection(set(dataid_list[i]))
                dataid_list = list(dataid)
            else:
                dataid = dataid_list[0]
                for i in range(1, len(dataid_list), 1):
                    dataid = set(dataid).union(set(dataid_list[i]))
                dataid_list = list(dataid)
            return dataid_list
        return dataid_list[0]

    @staticmethod
    def get_datetime_obj(timestr):
        return datetime.datetime.strptime(timestr, '%Y-%m-%d %H:%M:%S')

    @staticmethod
    def filter_time(df, from_, to_):
        datetime.datetime.strptime(from_, '%Y-%m-%d')
        datetime.datetime.strptime(to_, '%Y-%m-%d')
        start = from_
        end = to_
        mask = (df['time'] > start) & (df['time'] <= end)
        return df.loc[mask]

    @staticmethod
    def filter_by(df, **kwargs):
        '''
        Filter dataframe by city, dataid, time
        @param df: dataframe returned from query 
        @param Keyword Arguments:
                                 city (list, string): ['newyork', 'austin', 'california'] || 'newyork'
                                 dataid (list, string): ['5679', '5989'] || '5679'
                                 from_ (string): datetime string '%Y-%m-%d'
                                 to_ (string): datetime string '%Y-&m-&d'
        @return: filtered dataframe or empty list
        '''
        if 'from_' in kwargs and 'to_' not in kwargs:
            raise KeyError('[to_] argument is missing')
        if 'to_' in kwargs and 'from_' not in kwargs:
            raise KeyError('[from_] argument is missing')

        if 'city' in kwargs:
            if isinstance(kwargs.get('city'), list):
                df = df.loc[df['city'].isin(kwargs.get('city')), :]
            else:
                df = df.loc[df['city'] == kwargs.get('city'), :]

        if 'dataid' in kwargs:
            if isinstance(kwargs.get('dataid'), list):
                df = df.loc[df['dataid'].isin(kwargs.get('dataid')), :]
            else:
                df = df.loc[df['dataid'] == kwargs.get('dataid'), :]
        
        if 'from_' in kwargs:
            df = QueryDb.filter_time(df, kwargs.get('from_'), kwargs.get('to_'))

        if df.shape[0] == 0:
            return []
        return df

    def query_df(self, query_param, time_collection, **kwargs):
        '''
        @param query_param: list of query parameters
        @param time_collection: name of time collection in database -- ['time_(1_min|15_min)']
        @return: dataframe that matches query description
        @param Keyword Arguments:
                                  db_op (string): 'AND' || 'OR',
                                  year_tol (number): house_construction_year + tol,
                                  total_pv_tol (number): total_amount_pv + tol,
                                  total_sqft_tol (number): total_square_foot + tol,
                                  first_floor_sqft_tol (number): first_floor_square_footage + tol,
                                  second_floor_sqft_tol (number): second_floor_square_footage + tol
                                  third_floor_sqft_tol (number): third_floor_square_footage + tol
                                  half_floor_sqft_tol (number): half_floor_square_footage + tol,
                                  lower_level_sqft_tol (number): lower_level_square_footage + tol
        @return: dataframe or empty list
        '''
        assert len(query_param) > 0, 'must have at least one query item'

        if not isinstance(query_param, list):
            raise TypeError('query_param arguments should be type list')

        if len(query_param) == 1 and isinstance(query_param[0], dict):
            raise ValueError('at least one non-dictionary metadata parameter must be included in query')

        if len(query_param) > 1:
            count = 0
            for query in query_param:
                if isinstance(query, str) and query not in self.building_types:
                    count += 1
            if count == 0:
                raise ValueError('at least one non-dictionary metadata parameter must be included in query')

        if 'city' not in kwargs.keys():
            raise KeyError('[city] missing from arguments')

        if not isinstance(kwargs['city'], list):
            raise TypeError('[city] argument must be of type list')

        if len(query_param) > 1 and 'db_op' not in kwargs.keys():
            raise KeyError('multiple query parameters needs [dp_op]=[AND, OR]')

        for query in query_param:
            if not isinstance(query, str) and not isinstance(query, dict):
                raise TypeError('arguments to query_param must be of type string or dictionary')

        if 'pv' in query_param:
            raise ValueError('[pv] should be either [solar] or [solar2] or both')

        if len(query_param) == 1 and query_param[0] in self.building_types:
            raise ValueError('bulding type cannot be the only query parameter')

        #check for tol parameter in query
        tol_check = self.check_tol_param(query_param, kwargs)
        if not tol_check[0]:
            raise KeyError(tol_check[1] + ' is missing from Keyword Arguments')

        data_obj = []

        metadata = self.read_meta(time_collection)

        # return empty list if no value found for single query
        if len(query_param) == 1:
            if isinstance(query_param[0], dict):
                for key in query_param[0].keys():
                    if len(metadata[key]) == 0:
                        return data_obj
            else:
                if len(metadata[query_param[0]]) == 0:
                    return data_obj
        
        dataid_list = self.get_dataids(query_param, metadata, kwargs)

        cities = []

        cities = kwargs.get('city')

        if kwargs.get('city')[0] == 'all':
            cities = ['newyork', 'california', 'austin']

        for city in tqdm(cities, desc='fetching data from local server', ncols=100):
            data_file = self.read_file(time_collection, city)
            if not any(str(dataid) in data_file[city] for dataid in dataid_list):
                continue
            else:
                for dataid in dataid_list:
                    dataid = str(dataid)
                    if dataid in data_file[city].keys():
                        data_values = data_file[city][dataid]
                        for data_value in data_values:
                            value = {}
                            value['dataid'] = dataid
                            value['time'] = QueryDb.get_datetime_obj(data_value.get('time'))
                            value['city'] = city

                            for query in query_param:
                                if isinstance(query, str):
                                    if query in self.file_params:
                                        value[query] = data_value.get(query)
                                if query == 'nest':
                                    value['air1'] = data_value.get('air1')
                            data_obj.append(value)

        if len(data_obj) != 0:
            data_obj = pd.DataFrame(data_obj)
            print(data_obj.info(verbose=False, memory_usage='deep'))   

        return data_obj, dataid_list


        

        
        
        
            
        

